/* 	MemberTemplate.java
 *	Copyright (c) 2004, Brains2B.org
 *	
 *	Created by: Dennis Groenendijk
 *	Created on: Sept 19, 2002
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that  * the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this list of conditions and 
 * the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
 * the following disclaimer in the documentation and/or other materials provided with the distribution. 
 * 3. The name of the author may not be used to endorse or promote products derived from this software 
 * without specific prior written permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/* @#History
 * [When]		[Who]				[What]
 * 10/28/2002	dennis@brains2b.nl	Moved methods to Template interface and TemplateHelper
 * 11/06/2002		"		"		Add literal option to intialization
 */
package org.brains2b.typewriter;

import java.io.PrintWriter;
import java.lang.reflect.Modifier;

import org.brains2b.util.StringHelper;

/** Template for creating a Member in a class
 *
 * @author <A HREF="MAILTO:dennis@brains2b.nl">dennis@brains2b.nl</A>
 * @version 0.10 [12-9-2002]
 */
public class MemberTemplate implements Template {

	private ClassTemplate m_parent;
	private int m_modifiers;
	private String m_type;
	private String m_name;
	private Object m_initial;
	private boolean m_literal;

	/**
	 * Constructor
	 * <p>Make a private member with the given name and type
	 * @param type String, representation of the return value
	 * @param name String, the name of the method
	 */
	public MemberTemplate(String type, String name) {
		this(Modifier.PRIVATE,type,name,null,false);
	}

	/**
	 * Constructor
	 * <p>Make a private member with the given name and type
	 * @param type String, representation of the return value
	 * @param name String, the name of the method
	 * @param initialValue Object, the initial value for this method
	 * @param literal boolean, use the initial value as literal instead of a String
	 */
	public MemberTemplate(String type, String name, Object initialValue, boolean literal) {
		this(Modifier.PRIVATE,type,name,initialValue,literal);
	}

	/**
	 * Constructor
	 * <p>Make a member with the given name and type and with the given
	 * access.
	 * @param modifiers int, sum of the modifiers to use for this method
	 * @param type String, representation of the return value
	 * @param name String, the name of the method
	 */
	public MemberTemplate(int modifiers, String type, String name)  {
		this (modifiers,type,name,null,false);
	}
	
	/**
	 * Constructor
	 * <p>Make a member with the given name and type and with the given
	 * access and initial value.
	 * @param modifiers int, sum of the modifiers to use for this method
	 * @param type String, representation of the return value
	 * @param name String, the name of the method
	 * @param initialValue Object, the initial value for this method
	 * @param literal boolean, use the initial value as literal instead of a String
	 */
	public MemberTemplate(int modifiers, String type, String name,Object initialValue, boolean literal)  {
		super();
		if (type==null || type.length()==0) {
			throw new TypeWriterException("Member type must be supplied");	
		}
		if (name==null || name.length()==0) {
			throw new TypeWriterException("Member name cannot be empty");
		}
		m_modifiers=modifiers;
		m_type=type;
		m_name=name;
		m_initial=initialValue;
		m_literal=literal;
	}
	
	/* (non-Javadoc)
	 * @see org.brains2b.typewriter.Template#write(java.io.PrintWriter, int)
	 */
	public void write(PrintWriter pw, int indent) {
		indent=indent*4;
		if (Modifier.isStatic(m_modifiers) && Modifier.isFinal(m_modifiers) && m_initial ==null) {
			throw new TypeWriterException("Cannot make a final static member without an initial value");
		}
		pw.print(StringHelper.leftPad(""," ",indent));
		pw.print(TypeWriterHelper.getModifiers(m_modifiers));
		pw.print(m_type);
		pw.print(" ");
		pw.print(m_name);
		if (m_initial!=null) {
			pw.print(" = ");
			if (!m_literal && m_type.equals("String")) {
				pw.print('\"');
				pw.print(m_initial);
				pw.print('\"');
			} else {
				pw.print(m_initial.toString());
			}
		}
		pw.print(";");
		
	}
	/**
	 * get the initial value for this member
	 * @return Object
	 */
	public Object getInitialValue() {
		return m_initial;
	}

	/**
	 * set the initial value for this member
	 * @param initialValue Object
	 */
	public void setInitialValue(Object initialValue) {
		m_initial = initialValue;
	}
	
	/**
	 * set the initial value for this member as literal
	 * <p>setting literal to false is same as <code>setInitialValue(Object)</code>
	 * @param initialValue Object, initial value
	 * @param literal boolean, true if this is a literal
	 */
	public void setInitialValue(Object initialValue, boolean literal) {
		m_initial=initialValue;
		m_literal=literal;
	}

	/* (non-Javadoc)
	 * @see org.brains2b.typewriter.Template#getParent()
	 */
	public Template getParent() {
		return m_parent;
	}

	/* (non-Javadoc)
	 * @see org.brains2b.typewriter.Template#setParent(org.brains2b.typewriter.Template)
	 */
	public void setParent(Template parent) {
		m_parent =(ClassTemplate) parent;
	}

	/* (non-Javadoc)
	 * @see org.brains2b.typewriter.Template#addImport(java.lang.String)
	 */
	public void addImport(String name) {
		m_parent.addImport(name);
	}

	/** 
	 * @see org.brains2b.typewriter.Template#getObjectName(java.lang.String)
     * @deprecated use TypeWriterHelper.getClassName(String)
	 */
	public String getObjectName(String name) {
		return TypeWriterHelper.getClassName(name);
	}

}
