/* 	MethodTemplate.java
 *	Copyright 	(c) 2002, Brains2B.org
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided that
 * the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this list of conditions and 
 * the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
 * the following disclaimer in the documentation and/or other materials provided with the distribution. 
 * 3. The name of the author may not be used to endorse or promote products derived from this software 
 * without specific prior written permission. 
 * 
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/* @#History
 * [When]		[Who]				[What]
 * 09/05/2002	dennis@brains2b.nl	add Exceptions to methods
 * 10/28/2002	dennis@brains2b.nl	Moved methods to Template interface and TemplateHelper
 * 11/27/2002	dennis@brains2b.nl	Implement exception ones.
 * 19-02-2003	dennis@brains2b.nl	Implemented comparable to sort methods in the Target class
 */
package org.brains2b.typewriter;

import java.io.PrintWriter;
import java.lang.reflect.Modifier;
import java.util.TreeSet;

import org.brains2b.util.StringHelper;

/** Template for creating a Method in a class
 *
 * @author <A HREF="MAILTO:dennis@brains2b.nl">dennis@brains2b.nl</A>
 * @version 0.10 [05-09-2002]
 * @see org.brains2b.typewriter.ClassTemplate#createMethod(int, String, String, String[])
 * @see org.brains2b.typewriter.ClassTemplate#createMethod(int, String, String, String[][])
 * @see org.brains2b.typewriter.ClassTemplate#createMethod(String, String, String[])
 * @see org.brains2b.typewriter.ClassTemplate#createMethod(String, String[])
 * @see org.brains2b.typewriter.ClassTemplate#createConstructor(String[])
 * @see org.brains2b.typewriter.ClassTemplate#createConstructor(String[][])
 */
public class MethodTemplate extends ContentTemplate implements Comparable {
	
	private String m_name;
	private String m_returnType;
	private String[] m_params;
	private String[] m_paramValues;
	private TreeSet m_exceptions;
	private int m_modifiers;
	private boolean m_constructor;
	
	/**
	 * Constructor
	 * <p>This constructor will use number "arg" arguments as parameter names, i.e. arg0, arg1
	 * @param parent ClassTemplate, the parent which contains this element
	 * @param modifiers int, sum of Modifier constants which define the access to this method
	 * @param name String, name of the method
	 * @param returnType String, representation of the return type of this method i.e. "int", "String"
	 * <p>If null and this method is not defined as <class>setConstructor(boolean)</class> void will
	 * be used
	 * @param params String[], String representation of parameter types for this method ie. {"int","String"}
	 * @see #setConstructor(boolean)
	 */
	public MethodTemplate(ClassTemplate parent, int modifiers, String name, String returnType, String[] params) {
		super();
		setParent(parent);
		m_name=name;
		m_params=params;
		m_returnType=returnType;	
		m_modifiers=modifiers;
	}
	
	/**
	 * Constructor
	 * @param parent ClassTemplate, the parent which contains this element
	 * @param modifiers int, sum of Modifier constants which define the access to this method
	 * @param name String, name of the method
	 * @param returnType String, representation of the return type of this method i.e. "int", "String"
	 * <p>If null and this method is not defined as <class>setConstructor(boolean)</class> void will
	 * be used
	 * @param params String[][], String representation of parameter types and parameter names for this method ie. {{"int","String"},{"id","name"}}
	 * @see #setConstructor(boolean)
	 */
	public MethodTemplate(ClassTemplate parent, int modifiers, String name, String returnType, String[][] params) {
		super();
		setParent(parent);
		m_name=name;
		m_params=params[0];
		m_paramValues=params[1];
		m_returnType=returnType;	
		m_modifiers=modifiers;
	}

	/**
	 * add a throws clause to the method
	 * <p>Call this for every Exception this method can throw
	 * @param exName String, qualified name of the Exception 
	 */
	public void addException(String exName) {
		String name=TypeWriterHelper.getClassName(exName);
		if (m_exceptions==null) {
			m_exceptions=new TreeSet();
		}
		m_exceptions.add(name);
		
	}
	
	/**
	 * invoke a method from another instance of an object
	 * TODO This method needs to change
	 * @param instance String, instance of an object
	 * @param methodName String, the method to be called from the instance
	 * @param value Object, a String or a Literal object
	 */
	public void invokeMethod(String instance, String methodName,Object value) {
		invokeMethod(instance,methodName,new Object[]{value});
	}
	
	/**
	 * invoke a method from another instance of an object
	 * TODO implementation needs to change to only use Strings if explicitly defined as Literal
	 * @param instance String, instance of an object
	 * @param methodName String, the method to be called from the instance
	 * @param args Object[], list of arguments to pass to the the method 
	 */
	public void invokeMethod(String instance, String methodName,Object[] args) {
		StringBuffer sb=new StringBuffer();
		sb.append(instance).append(".").append(methodName);
		if (args==null) {
			sb.append("();");
		} else {
			sb.append("( ");
			for (int i=0;i<args.length;i++) {
				sb.append(String.valueOf(args[i]));
				
				if (i<args.length-1) {
					sb.append(", ");
				} 
			}
			sb.append(" );");
		}
		m_content.add(sb.toString());
	}
	
	/* (non-Javadoc)
	 * @see org.brains2b.typewriter.Template#write(java.io.PrintWriter, int)
	 */
	public void write(PrintWriter pw, int indent) {
		String spc=StringHelper.leftPad(""," ",indent*4);
		pw.print(spc);
		pw.print(getModifiers());
		if (!isConstructor()) {
			pw.print(m_returnType==null?"void":m_returnType);
			pw.print(" ");
		}
		pw.print(m_name);
		if (m_params==null) {
			pw.print("() ");
		} else {
			pw.print("(");
			for (int i=0;i<m_params.length;i++) {
				pw.print(m_params[i]);
				if (m_paramValues==null) {
					pw.print( " arg" + i);
				} else {
					pw.print(" " + m_paramValues[i]);
				}
				
				if (i<m_params.length-1) pw.print(", ");
			}
			pw.print(") ");
		}
		if(m_exceptions!=null) {
			pw.print("throws ");
			for (int i=0;i<m_exceptions.size();i++) {
				pw.print((String) m_exceptions.toArray()[i]);
				if (i+1 < m_exceptions.size()) {
					pw.print(",");
				}
				pw.print(" ");
			}
		}
		pw.println("{");
		
		super.write(pw,indent+1);
		pw.print(spc);
		pw.println("}");
		 
	}
	
	/**
	 * get the String representation of the modifiers used by this method
	 * @return String
	 * @see #MethodTemplate(ClassTemplate, int, String, String, String[])
	 * @see #MethodTemplate(ClassTemplate, int, String, String, String[][])
	 */
	private String getModifiers() {
		return TypeWriterHelper.getModifiers(m_modifiers);
	} 
	
	/**
	 * check if this method is a constrcutor instead of a method
	 * @return boolean, true if this MethodTemplate contains a constructor,
	 * false otherwise
	 */
	public boolean isConstructor() {
		return m_constructor;
	}

	/**
	 * set this Methodtemplate to be a constructor instead of a normal
	 * method
	 * @param b boolean, true if this is a constructor, false otherwise
	 */
	public void setConstructor(boolean b) {
		m_constructor = b;
	}
	
    /**
     * get the name of this method
     * @return String
     */
    public String getName() {
        return m_name;
    }

	/**
	 * compare this object to another MethodTemplate
	 * <UL>Compares on the following properties
	 * <LI>Modifiers, public goes for protected goes for private
	 * <LI>Name of the method
	 * <LI>Number of arguments to this method, smaller number before
	 * larger
	 * @see java.lang.Comparable#compareTo(Object)
	 */
	public int compareTo(Object o) {
		int result=-1;
		if (o instanceof MethodTemplate) {
			MethodTemplate mth2=(MethodTemplate) o;
			result=getCompareValue(this)-getCompareValue(mth2);			
			if (result==0) { result=this.getName().compareTo(mth2.getName());}
			if (result==0) {
				if (this.m_params!=null) {
					if (mth2.m_params!=null) {
						result=this.m_params.length-mth2.m_params.length;
					} else {
						result=1;
					}
				}
			}
		} else {
			result=1;
		}
		return result;
	}

	/**
	 * get the sort order for the modifiers and constructor
	 * for this method
	 * @param mth MethodTemplate
	 * @return int, sortable value for the modifiers of this method 
	 */
	private int getCompareValue(MethodTemplate mth) {
		int total=0;
		if (!mth.isConstructor()) total+=8;
		if (!Modifier.isStatic(mth.m_modifiers)) total+=4;
		if (Modifier.isProtected(mth.m_modifiers)) total+=2;
		if (Modifier.isPrivate(mth.m_modifiers)) total+=1;
		
		return total;
	}
	

}
